/* File : icod.c

   Simon Drabble	02/05/99
   madlather@syspac.com

   (C) 1998 - 1999  Simon Drabble

   This program and its associated code are released under
   the Gnu Public Licence (GPL). This basically states you
   are free to modify and distribute the software provided
   you do not prevent anyone else doing the same.        
   For a full copy of the licence, see www.gnu.org/copyleft/gpl.html

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "global.h"
#include "icod.h"


BYTE *global_colour_map = NULL;
char *global_file_name = NULL;
const char *ok_chars = "!@#$%^&*(){}[]|+=-,.<>/?;:abcdefghjiklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";


#define USAGE	"Usage: icod file1.ico [file2.ico .. filen.ico]"

int main(int argc, char *argv[])
{
  int i;

#ifdef DEBUG
  fprintf(stderr, "BYTE (%d)\nWORD (%d)\nDWORD (%d)\nLONG (%d)\nICONDIR (%d)\nICONDIRENTRY (%d)\n",
          sizeof(BYTE), sizeof(WORD), sizeof(DWORD), sizeof(LONG),
          sizeof(ICONDIR), sizeof(ICONDIRENTRY));

  fprintf(stderr, "BITMAPINFOHEADER (%d)\nRGBQUAD (%d)\n",
          sizeof(BITMAPINFOHEADER), sizeof(RGBQUAD));
#endif

  if (argc == 1) {
    printf("%s\n", USAGE);
    exit(0);
  }

  for (i = 1; i < argc; ++i) 
    load_icon(argv[i]);

  return 0;
  
}


int init_colour_map(ICONIMAGE *ii)
{
  register int i, j, k;
  int cpp = 2,
      bd = ii->icHeader.biBitCount,
      bc = 1 << bd;

  long size = bc * cpp;

  int b = strlen(ok_chars);
  
  free(global_colour_map);
  CREATE(global_colour_map, char, size + 1);

  memset(global_colour_map, 0, size);

  for (i = 0; i < b; ++i)
    for (j = 0; j < b; ++j) {
      k = (i*b+j) * cpp;
      if (k < size) {
        *(global_colour_map + k) = ok_chars[j];
        *(global_colour_map + k + 1) = ok_chars[i];
      } else
        return 0;
    }

}

  

int load_icon(char *fname)
{
  FILE *iconf;

  if (!(iconf = fopen(fname, "r b"))) 
    return -1;

  free(global_file_name);
  CREATE(global_file_name, char, strlen(fname) + 1);
  strcpy(global_file_name, fname);

  if ((strrchr(global_file_name, '.')))
    *(strrchr(global_file_name, '.')) = '\0';
  
  load_icon_file(iconf);
}



int load_icon_file(FILE *iconf)
{  
  ICONDIR *idir, temp;

  int i, tot = 0;


  CREATE(idir, ICONDIR, 1);

  memset(idir, 0, sizeof(ICONDIR));

  tot += fread(&idir->idReserved, sizeof(WORD), 1, iconf);
  tot += fread(&idir->idType, sizeof(WORD), 1, iconf);
  tot += fread(&idir->idCount, sizeof(WORD), 1, iconf);
  tot *= sizeof(WORD);


  CREATE(idir->idEntries, ICONDIRENTRY, idir->idCount); 
  tot = fread(idir->idEntries, sizeof(ICONDIRENTRY), idir->idCount, iconf);
  printf("\n%s: %d icons.\n", global_file_name, idir->idCount);


  for (i = 0; i < idir->idCount; ++i)
    load_icon_image(iconf, i, idir);

}


int load_icon_image(FILE *inf, int i, ICONDIR *idir)
{ 
  ICONIMAGE *ii;
  int tot = 0;
  size_t nc;
  size_t c;
  s_xpm *xpm;

  CREATE(ii, ICONIMAGE, 1);
  memset(ii, 0, sizeof(*ii));

  /* Am I really doing this? */
  fseek(inf, idir->idEntries[i].dwImageOffset, SEEK_SET);

  tot = fread(&ii->icHeader, sizeof(BITMAPINFOHEADER), 1, inf);

  dump_iconimage(ii);

  nc = 1 << ii->icHeader.biBitCount;
  if (ii->icHeader.biBitCount != 24) {
    CREATE(ii->icColours, RGBQUAD, nc);
    tot = fread(ii->icColours, sizeof(RGBQUAD), nc, inf);
    c = (ii->icHeader.biWidth * ii->icHeader.biHeight / 2) /
         (8 / ii->icHeader.biBitCount);

  } else {
/*
    CREATE(ii->icColours, RGBQUAD,
           ii->icHeader.biWidth * ii->icHeader.biHeight / 2);
*/
    ii->icColours = NULL;
    c = (ii->icHeader.biWidth * ii->icHeader.biHeight / 2) * 3;
    
  }

  CREATE(ii->icXOR, BYTE, c);
  tot = fread(ii->icXOR, sizeof(BYTE), c, inf);


  /* really * 2 / 8 */
/*
  c = ii->icHeader.biWidth * ii->icHeader.biHeight >> 2;
  CREATE(ii->icAND, BYTE, c); 
  tot = fread(ii->icAND, sizeof(BYTE), c, inf);
*/

  CREATE(xpm, s_xpm, 1); 
  xpm->cpp = 2;

  if (ii->icHeader.biBitCount == 24) {
    process_24bit(ii, xpm);
  } else {
    init_colour_map(ii);
    ico2xpm(ii, xpm); 
    write_xpm(ii, xpm);
  }


  free(xpm->data);  
  free(xpm);
  free(ii->icAND);
  free(ii->icXOR);
  free(ii->icColours);
  free(ii);
  
  return 0;
}



int process_24bit(ICONIMAGE *ii, s_xpm *xpm)
{
  register int i, j, k, l, m, n, p, q, r;
  int b = strlen(ok_chars);
  int mult = 3 + xpm->cpp;
  int size = (ii->icHeader.biWidth * ii->icHeader.biHeight / 2) * mult;
  int offset;

  struct s_rgb {
    BYTE red, grn, blu;
  } rgb;

  FILE *xpmf;
  char *fname;

  CREATE(fname, char, strlen(global_file_name) + 64);
  strcpy(fname, global_file_name);
  
  xpm->width = ii->icHeader.biWidth;
  xpm->height = ii->icHeader.biHeight / 2;

  CREATE(xpm->data, char, (xpm->width * xpm->height * xpm->cpp));

  

  /* Build the colour_table */
  
  CREATE(global_colour_map, char, size);
  memset(global_colour_map, 0, size);

  n = p = q = 0;

  l = ii->icHeader.biHeight / 2;
  r = ii->icHeader.biWidth * 3;

  for (i = 0; i < l; ++i)
    for (j = 0; j < r; j += 3) {
      offset = i * r + j;
      rgb.blu = *(ii->icXOR + offset);
      rgb.grn = *(ii->icXOR + offset + 1);
      rgb.red = *(ii->icXOR + offset + 2);

      for (k = 0; k < n; k += mult) 
        if (!memcmp(&rgb, global_colour_map + k + xpm->cpp,
                    sizeof(struct s_rgb)))
            break;

      if (k >= n) {
        n += mult;
        *(global_colour_map + k) = ok_chars[p];
        *(global_colour_map + k + 1) = ok_chars[q];
        if (++p >= b) {
          p = 0; ++q;
        }
        if (q >= b) {
          printf("Ahaarrr matey, ran out of colours! Shiver me timbers!\n");
          return -1;
        }
        *(global_colour_map + k + xpm->cpp) = rgb.red;
        *(global_colour_map + k + xpm->cpp + 1) = rgb.grn;
        *(global_colour_map + k + xpm->cpp + 2) = rgb.blu;
      }

      m = i * xpm->width + (j / 3);
      *(xpm->data + (m * xpm->cpp)) = *(global_colour_map + k);
      *(xpm->data + (m * xpm->cpp) + 1) = *(global_colour_map + k + 1);
    }

  xpm->num_cols = n / mult;

  sprintf(fname, "%s%02dX%02d_%d.xpm",
          global_file_name, xpm->width, xpm->height, ii->icHeader.biBitCount);

  if (!(xpmf = fopen(fname, "w")))
    return -1; 

  free(fname);

  write_xpm_hdr(xpm, xpmf);

  k = (xpm->num_cols * mult);
  for (i = 0; i < k; i += mult)
    fprintf(xpmf, "\"%c%c\tc #%02x%02x%02x\",\n",
            *(global_colour_map + i), *(global_colour_map + i + 1),
            *(global_colour_map + i + xpm->cpp),
            *(global_colour_map + i + xpm->cpp + 1),
            *(global_colour_map + i + xpm->cpp + 2));

  write_xpm_data(xpm, xpmf);

  fclose(xpmf);
  return 0;
}




int ico2xpm(ICONIMAGE *ii, s_xpm *xpm)
{
  register int i, j, k, l, m, n, p, q;
  BYTE pixel, byte;

  k = ii->icHeader.biBitCount;
  xpm->width = ii->icHeader.biWidth;
  xpm->height = ii->icHeader.biHeight / 2;
  xpm->num_cols = (1 << k);

  CREATE(xpm->data, char, xpm->width * xpm->height * xpm->cpp);



  p = 8 / k;
  q = (1 << k) - 1;
  
  for (i = 0; i < xpm->height; ++i) 
    for (j = 0; j < xpm->width; ++j) {

      int offset = (i * xpm->width + j);

       /* This is where we use the bit-depth info */
      byte = *(ii->icXOR + offset / p);
      for (l = 8 - k, n = 0; l >= 0; l -= k, ++n) {
        m = offset + n;
        pixel = (byte >> l) & q;

        *(xpm->data + m * xpm->cpp) =
          global_colour_map[pixel * xpm->cpp];

        *(xpm->data + m * xpm->cpp + 1) =
          global_colour_map[pixel * xpm->cpp + 1];
      }

    }

  return 0;
}


int write_xpm_hdr(s_xpm *xpm, FILE *outf)
{
  fprintf(outf, "/* XPM */\n");
  fprintf(outf, "char * image_data[] = {\n");
  fprintf(outf, "\"%d  %d  %d  %d\",\n",
          xpm->width, xpm->height, xpm->num_cols, xpm->cpp);
}


int write_xpm(ICONIMAGE *ii, s_xpm *xpm)
{
  register int i, j;
  int base = 0;
  FILE *xpmf;
  char *fname;

  CREATE(fname, char, strlen(global_file_name) + 64);
  strcpy(fname, global_file_name);
  
  sprintf(fname, "%s%02dX%02d_%d.xpm",
          global_file_name, xpm->width, xpm->height, ii->icHeader.biBitCount);

  if (!(xpmf = fopen(fname, "w")))
    return -1; 

  free(fname);

  write_xpm_hdr(xpm, xpmf);
  write_colour_table(ii, xpmf);

  write_xpm_data(xpm, xpmf);

  fclose(xpmf);
  return 0;
}



int write_xpm_data(s_xpm *xpm, FILE *xpmf)
{
  register int i, j, offset;

  /* For some reason data is stored upside down.. */
  for (i = xpm->height - 1; i >= 0; --i) {
    fprintf(xpmf, "\"");
    for (j = 0; j < xpm->width; ++j) {
      offset = i * xpm->width + j;
      fprintf(xpmf, "%1c%1c",
              *(xpm->data + offset * xpm->cpp),
              *(xpm->data + offset * xpm->cpp + 1));
    }
    fprintf(xpmf, "\",\n");
  }
  fprintf(xpmf, "};\n\n");

  return 0;
}


int dump_andmask(ICONIMAGE *ii)
{
  register int i, j, k;
  int base = 0;
  for (i = 0; i < ii->icHeader.biHeight ; i += 2) {
    for (j = 0; j < ii->icHeader.biWidth; j += 8) {
      for (k = 7; k >= 0; --k) {
        base = *(ii->icAND + ((i * ii->icHeader.biWidth + j) / 8)) & (1 << k);
        fprintf(stderr, "%c",
                base ? '1' : '0');
      }
    }
    fputc('\n', stderr);
        
  }
  fputc('\n', stderr);
} 
 

int write_colour_table(ICONIMAGE *ii, FILE *outf)
{
  long i; 

  for (i = 0; i < (long) (1 << ii->icHeader.biBitCount); ++i) {
    fprintf(outf, "\"%1c%1c\tc #%02x%02x%02x\",\n",
            global_colour_map[i * 2], 
            global_colour_map[i * 2 + 1], 
            ii->icColours[i].red,
            ii->icColours[i].green,
            ii->icColours[i].blue);
  }
  fputc('\n', outf);

}
  


int dump_iconimage(ICONIMAGE *ii)
{
  printf("ICON: %2ldx%2ld %2ldbpp Sz:%ld\n", 
         ii->icHeader.biWidth, ii->icHeader.biHeight / 2,
         ii->icHeader.biBitCount,
         ii->icHeader.biSizeImage);

}
  
int dump_icondir(ICONDIR *idir)
{
  int i;
  fprintf(stderr, "%4ld : Reserved\n%4ld : Type\n%4ld : Count\n",
          idir->idReserved, idir->idType, idir->idCount);
  dump_icondirent(0, idir);
  
}


int write_icondir(char *name, ICONDIR *idir)
{
  FILE *fout;
  fout = fopen(name, "w b");
  fwrite(idir, sizeof(ICONDIR), 1, fout);
  fclose(fout);
}
  
int dump_icondirent(int i, ICONDIR *idir)
{ 

  fprintf(stderr, "%2d: Wd(%d) Ht(%d) CC(%d) Bc(%d) Sz(%04lx) Of(%04lx)\n", i,
          idir->idEntries[i].bWidth,
          idir->idEntries[i].bHeight,
          idir->idEntries[i].bColourCount,
          idir->idEntries[i].wBitCount,
          idir->idEntries[i].dwBytesInRes,
          idir->idEntries[i].dwImageOffset);
  
}

int dump_xormask(ICONIMAGE *ii, size_t c)
{
  int i;
  fprintf(stderr, "Dumping %d bytes.\n", c);
  for (i = 0; i < c; ++i) {
    if (!(i % 48)) fputc('\n', stderr);
    if (!(i % 3)) fputc(' ', stderr);
    fprintf(stderr, "%02x", *(ii->icXOR + i));
  }
  fputc('\n', stderr);
}
